import QtQuick 2.12
import QtQuick.Controls 2.12

import NERvGear 1.0 as NVG
import NERvGear.Controls 1.0
import NERvGear.Templates 1.0 as T

T.Widget {
    id: widget

    readonly property QtObject styleSettings: NVG.Settings.makeMap(settings, "style")

    readonly property var itemMetrics: {
        switch (styleSettings.size) {
        case 3: return { height: 24, gap: 3, font: 14 };
        case 2: return { height: 23, gap: 2, font: 13 };
        case 1: return { height: 22, gap: 2, font: 12 };
        case 0:
        default: break;
        }
        return { height: 20, gap: 2, font: 11 };
    }

    readonly property real unitOpacity: 0.5
    readonly property int unitSize: itemMetrics.height
    readonly property int unitGap: itemMetrics.gap

    readonly property color colorBGH: "#666"
    readonly property color colorBGP: "#333"
    readonly property color btnColorFG: "#FFF"
    readonly property color btnColorBG1: "#000"
    readonly property color btnColorBG2: "#FF9326"

    property font textFont

    textFont {
        pixelSize: itemMetrics.font
        family: "Source Han Sans"
    }

    readonly property url channelUrl: settings.url || "http://www.gpbeta.com/feed/"
    property bool channelError
    property int channelInterval
    property string channelTitle
    property string channelBuild
    property var channelDate: null
    property var xhrCurrent: null

    readonly property var itemFilter: {
        const script = settings.filter ?? "";

        if (script) {
            try {
                const filter = Function("element", "item", "'use strict';\n" + script);
                return function (element, item) {
                    try {
                        return filter(element, item);
                    } catch (err) {
                        item.title = formatScriptError(err);
                        console.warn(item.title);
                    }
                    return true;
                }
            } catch (err) {
                const str = formatScriptError(err);
                console.warn(str);
                return function (element, item) {
                    item.title = str;
                    return true;
                };
            }
        }

        return ()=>true;
    }

    title: channelTitle || qsTr("RSS Feed")
    solid: true
    enabled: !xhrCurrent

    implicitWidth: 200
    implicitHeight: 200

    menu: FeedWidgetMenu { }

    function formatScriptError(err) {
        if (!err)
            return "";

        if (!(err instanceof Error))
            return err.toString();

        if (err instanceof SyntaxError)
            return err.toString();

        return err.name + " @" + (err.lineNumber - 2) + ": " + err.message;
    }

    function fixMissingElementMethods(element) {
        // workaround for missing element methods in XMLHttpRequest implementation

        if (element.hasAttribute)
            return;

        const proto = Object.getPrototypeOf(element);

        proto.getElementsByTagName = function (tag) {
            const elements = [];
            for (let child = this.firstChild; child; child = child.nextSibling) {
                if (child.nodeType === 1 && child.tagName === tag)
                    elements.push(child);
            }
            return elements;
        };

        proto.hasAttributes = function () {
            return this.attributes?.length ? true : false;
        };

        proto.getAttribute = function (name) {
            return this.attributes?.[name]?.value ?? null;
        };

        proto.hasAttribute = function (name) {
            return this.attributes?.[name] ? true : false;
        };
    }

    function getElementByTagName(node, tag) {
        for(let child = node.firstChild; child; child = child.nextSibling) {
            if (child.nodeType === 1 && child.tagName === tag)
                return child;
        }
        return null;
    }

    function getInnerText(element) {
        var text = "";
        for (let child = element?.firstChild; child; child = child.nextSibling)
            text += child.wholeText;
        return text;
    }

    function getNodeDate(node) {
        return new Date(node ? getInnerText(node) : 0);
    }

    function initializeChannel(xhr) {
        const doc = xhr.responseXML.documentElement;
        const channel = getElementByTagName(doc, "channel");

        fixMissingElementMethods(doc);

        channel.getElementsByTagName("item").forEach(function (element) {
            const item = {
                date: getNodeDate(getElementByTagName(element, "pubDate")),
                title: getInnerText(getElementByTagName(element, "title")).trim(),
                link: getInnerText(getElementByTagName(element, "link")),
                guid: getInnerText(getElementByTagName(element, "guid")),
                highlight: false
            };
            element.findElementByTagName = getElementByTagName;
            if (item.link && item.guid && itemFilter(element, item))
                feedModel.append(item);
        });
        channelTitle = getInnerText(getElementByTagName(channel, "title"));
        channelBuild = getInnerText(getElementByTagName(channel, "lastBuildDate"));
        const ttlText = getInnerText(getElementByTagName(channel, "ttl"));
        if (ttlText) {
            let ttl = Number(ttlText);
            if (ttl >= 5 && ttl <= 120)
                channelInterval = ttl * 60000;
        }
    }

    function updateChannel(xhr) {
        const doc = xhr.responseXML.documentElement;
        const channel = getElementByTagName(doc, "channel");
        const build = getInnerText(getElementByTagName(channel, "lastBuildDate"));

        if (build === channelBuild)
            return;

        channelBuild = build;

        var insertPos = 0;

        channel.getElementsByTagName("item").forEach(function (element) {
            const guid = getInnerText(getElementByTagName(element, "guid"));

            if (!guid)
                return;

            // skip existed items
            for (let i = insertPos; i < feedModel.count; ++i) {
                const props = feedModel.get(i);
                if (props.guid === guid)
                    return;
            }

            const link = getInnerText(getElementByTagName(element, "link"));

            if (!link)
                return;

            const item = {
                date: getNodeDate(getElementByTagName(element, "pubDate")),
                title: getInnerText(getElementByTagName(element, "title")).trim(),
                link: link,
                guid: guid,
                highlight: true
            };

            if (!itemFilter(element, item))
                return;

            feedModel.insert(insertPos, item);
            ++insertPos;
        });

        // clear previous highlights
        if (insertPos) {
            for (let i = insertPos; i < feedModel.count; ++i)
                feedModel.setProperty(i, "highlight", false);
        }
    }

    function updateFeedModel(xhr) {
        try {
            if (feedModel.count)
                updateChannel(xhr);
            else
                initializeChannel(xhr);
            channelDate = new Date;
            channelError = false;
        } catch(err) {
            channelError = true;
            console.warn(err);
        }
    }

    function updateError(xhr) {
        channelError = true;
        console.warn("cannot retrieve RSS data: " + xhr.status + " " + xhr.statusText);
    }

    function feedChannel(url) {
        const xhr = new XMLHttpRequest();
        xhrCurrent = xhr;
        xhr.onerror = function () {
            if (xhrCurrent !== xhr)
                return;

            updateError(xhrCurrent);
            xhrCurrent = null;
        }
        xhr.onload = function () {
            if (xhrCurrent !== xhr)
                return;
            // BUG: null responseXML when redirect
            if (xhr.status >= 200 && xhr.status < 300) {
                updateFeedModel(xhrCurrent);
                xhrCurrent = null;
            } else {
                xhr.onerror();
            }
        }
        xhr.responseType = 'document';
        xhr.open("GET", url);
        xhr.send();
    }

    function reloadChannel() {
        channelDate = null;
        channelError = false;
        channelInterval = 0;
        channelTitle = "";

        feedView.positionViewAtBeginning();
        feedModel.clear();

        feedChannel(channelUrl);
    }

    // realod or Component.onCompleted
    onItemFilterChanged: {
        if (!xhrCurrent) {
            if (feedModel.count)
                reloadChannel();
            else
                feedChannel(channelUrl);
        }
    }

    Timer {
        running: !xhrCurrent
        interval: {
            if (channelError)
                return 5 * 60000;

            if (widget.settings.interval)
                return widget.settings.interval;

            if (channelInterval)
                return channelInterval;

            return 15 * 60000;
        }
        onTriggered: feedChannel(channelUrl)
    }

    FeedWidgetButton {
        id: messageArea
        anchors.top: parent.top
        anchors.left: parent.left
        anchors.right: refreshButton.left
        anchors.rightMargin: unitGap

        enabled: false
        font: widget.textFont
        text: {
            if (xhrCurrent)
                return qsTr("Now Loading...");
            var str = channelError ? qsTr("Fetch Error") : qsTr("RSS Updated");
            if (channelDate)
                str += " @ " + Qt.formatDateTime(channelDate, "hh:mm");
            return str;
        }

        contentItem: Text {
            text: messageArea.text
            color: widget.btnColorFG
            horizontalAlignment: Text.AlignHCenter
            verticalAlignment: Text.AlignVCenter
            style: Text.Outline
            styleColor: "transparent"
            font: messageArea.font
        }
    }

    FeedWidgetButton {
        id: refreshButton
        anchors.top: parent.top
        anchors.right: parent.right

        icon {
            name: "regular:\uf2f9"
            color: btnColorFG
            width: unitSize / 2
            height: width
        }

        contentItem: Icon { icon: refreshButton.icon }

        onClicked: feedChannel(channelUrl)
    }

    ListView {
        id: feedView
        anchors.top: refreshButton.bottom
        anchors.topMargin: unitGap * 2
        anchors.bottom: parent.bottom
        anchors.left: parent.left
        anchors.right: parent.right

        clip: true
        spacing: unitGap

        add: Transition {
            id: addTrans
            SequentialAnimation {
                PropertyAction { property: "opacity"; value: 0 }
                PauseAnimation {
                    duration: addTrans.ViewTransition.index * 50 + 250
                }
                NumberAnimation { property: "opacity"; from: 0; to: 1; duration: 250 }
            }
        }
        remove: Transition {
            id: removeTrans
            SequentialAnimation {
                PauseAnimation {
                    duration: {
                        const targets = removeTrans.ViewTransition.targetIndexes;
                        return Math.max(targets.length - removeTrans.ViewTransition.index, 0) * 50;
                    }
                }
                NumberAnimation { property: "opacity"; from: 1; to: 0; duration: 250 }
            }
        }
        displaced: Transition {
            NumberAnimation { property: "y"; duration: 250; easing.type: Easing.OutCubic }
        }

        model: ListModel { id: feedModel }
        delegate: FeedWidgetButton {
            id: control

            width: widget.width
            color: model.highlight ? widget.btnColorBG2 : widget.btnColorBG1
            text: model.title
            leftInset: timeRect.width + widget.unitGap * 2
            leftPadding: leftInset

            onClicked: {
                feedModel.setProperty(index, "highlight", false);

                const open = (widget.settings.externalBrowser ?? true) ?
                               Qt.openUrlExternally : NVG.SystemCall.openLink;
                Qt.callLater(open, model.link);
            }

            Rectangle {
                id: timeRect
                width: widget.unitSize * 2.1
                height: widget.unitSize
                color: widget.btnColorBG1

                opacity: control.hovered ? 0.7 : 0.35
            }

            Text {
                id: timeLabel
                anchors.fill: timeRect
                anchors.margins: widget.unitGap
                anchors.leftMargin: widget.unitGap + 1

                color: widget.btnColorFG
                style: Text.Outline
                styleColor: "transparent"
                horizontalAlignment: Text.AlignHCenter
                verticalAlignment: Text.AlignVCenter

                text: {
                    if (!channelDate || !date)
                        return text;

                    if (date.getTime() === 0)
                        return "-- : --";

                    // different year
                    if (channelDate.getFullYear() !== date.getFullYear())
                        return date.getFullYear();

                    // different day
                    if (channelDate.getMonth() !== date.getMonth() ||
                        channelDate.getDate() !== date.getDate())
                        return Qt.formatDate(date, "MM - dd");

                    return Qt.formatDateTime(date, "hh : mm");
                }

                font.family: "SAO UI"
                font.pixelSize: widget.textFont.pixelSize
            }

            contentItem: Text {
                text: control.text
                color: widget.btnColorFG
                style: Text.Outline
                styleColor: "transparent"
                leftPadding: widget.unitGap * 3
                rightPadding: leftPadding
                elide: Text.ElideRight
                verticalAlignment: Text.AlignVCenter

                font: widget.textFont
            }
        }
    }
}
